/*
 *
 * UXmit APRS Utility
 *
 * Steve, N7MKO
 * 
 * $Id: uxmit.c,v 1.2 2015/01/11 11:49:43 steve Exp $
 *
 */

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

char mycall[16];
char passcode[16];
char latitude[16];
char longitude[16];
char string_buf[128];
char aprs_comment[256];
char aprs_buf[256];
char packet[512];
char recv_buf[2048];
char send_buf[2048];

char *default_aprs_comment = "UXmit v0.6, FreeBSD";
char *destcall = "APZ146";

char aprs_table = '/';
char aprs_symbol = 'x'; 

int aprs_is = 0;
int verbose = 0;

#define CONFIGFILE "/usr/local/etc/uxmit.conf"

int aprs_is_connect(char *, int, char *);

static struct option longopts[] = {

	{"latlon", no_argument, NULL, 'l'},
	{"symbol", no_argument, NULL, 's'},
	{"aprs-is", no_argument, NULL, 'i'},
	{"verbose", no_argument, NULL, 'v'},
	{"help", no_argument, NULL, 'h'},
	{NULL, 0, NULL, 0}
};

void usage()
{
	fprintf(stderr, "usage: uxmit [options] <APRS Comment>\n");
	fprintf(stderr, "\t-l, --latlon <0000.ss:00000.ss>\n");
	fprintf(stderr, "\t-s, --symbol <sym:tab>\n");
	fprintf(stderr, "\t-c, --config <file>\n");
	fprintf(stderr, "\t-i, --aprs-is <server>\n");
	fprintf(stderr, "\t-h, --help\n");
	exit(1);
}

void parse_latlon(char *arg)
{
	printf("todo: parse latlon()\n");
}

void load_config_file(char *path)
{
	FILE *fp;
	char line[80];
	char key[40];
	char value[40];

	if (!(fp = fopen(path, "r"))) {

		perror(path);
		return;
	}

	while (fgets(line, sizeof(line), fp) != NULL) {
		
		if (line[0] == '#')
			continue;
		
		sscanf(line, "%s %s", key, value);

		// printf("debug: key: %s val: %s\n", key, value);

		if (strcmp("callsign", key) == 0)
			snprintf(mycall, sizeof(mycall), "%s", value);
		else if (strcmp("passcode", key) == 0)
			snprintf(passcode, sizeof(passcode), "%s", value);
		else if (strcmp("latitude", key) == 0)
			snprintf(latitude, sizeof(latitude), "%s", value);
		else if (strcmp("longitude", key) == 0)
			snprintf(longitude, sizeof(longitude), "%s", value);
		else if (strcmp("aprs_symbol", key) == 0)
			aprs_symbol = value[0];
		else if (strcmp("aprs_comment", key) == 0)
			snprintf(aprs_comment, sizeof(aprs_comment), "%s", line + sizeof("APRS_COMMENT"));
	}
	
	fclose(fp);
}

char *aprs_report()
{
	char *nl;
	char *comment = NULL;

	if ((strlen(latitude) == 0) || (strlen(longitude) == 0)) {
		
		fprintf(stderr, "missing lat/lon\n");
		exit(4);
	}

	if (strlen(aprs_comment) > 0)
		comment = aprs_comment;
	else
		comment = default_aprs_comment;

	if ((nl = strchr(comment, '\n')) != NULL)
		*nl = '\0';

        sprintf(string_buf, "!%s%c%s%c%s\n", latitude, aprs_table, longitude, aprs_symbol, comment);

	return string_buf;
}

int main(int argc, char *argv[])
{
	char *p;
	int n = 0;
	int c = 0;
	int help = 0;

	while ((c = getopt_long(argc, argv, "ihv", longopts, NULL)) != EOF) {

		switch (c) {

			case 'l':
				parse_latlon(optarg);
				break;

			case 'i':
				aprs_is = 1;
				break;

			case 'v':
				verbose++;
				break;

			case 'h':
			default:
				help++;
		}
	}

	if (help)
		usage();

	load_config_file(CONFIGFILE);

	if (strlen(mycall) == 0) {

		fprintf(stderr, "no callsign\n");
		exit(2);
	}

	/* aprs comment is read from: config file, command line, or program default */

	if (optind < argc) {

		p = aprs_comment;

		if (argv[optind][0] == '-')
			while ((c = getchar()) != EOF) { *p = c; p++; }

		else
			while (optind < argc)
				n += snprintf(aprs_comment + n, sizeof(aprs_comment) - n, "%s ", argv[optind++]);
	}
	
	sprintf(aprs_buf, "%s>%s,TCPIP*:%s", mycall, destcall, aprs_report());

	if (aprs_is)
		aprs_is_connect("northwest.aprs2.net", 14578, aprs_buf);
	else
		printf("%s", aprs_buf);
        	
	return 0;
}

/* 
 * routines from Linux AX.25 utils to validate AX.25 addresses
 *
 */

#define AX25_MAX_DIGIS 8
#define AX25_MAX_FRAME_SIZE 256

int ax25_validate(const char *call)
{
        char s[7];
        int n;

        for (n = 0; n < 6; n++) {
                s[n] = (call[n] >> 1) & 0x7F;
        }

        s[6] = '\0';

        if (strspn(s, "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ ") == 6)
                return 1;

        return 0;
}

int ax25_aton_entry(const char *name, char *buf)
{
        int ct   = 0;
        int ssid = 0;
        const char *p = name;
        char c;

        while (ct < 6) {
                c = toupper(*p);

                if (c == '-' || c == '\0')
                        break;

                if (!isalnum(c)) {
                        printf("invalid symbol in callsign '%s'\n", name);
                        return -1;
                }

                buf[ct] = c << 1;

                p++;
                ct++;
        }

        while (ct < 6) {
                buf[ct] = ' ' << 1;
                ct++;
        }

        if (*p != '\0') {
                p++;

                if (sscanf(p, "%d", &ssid) != 1 || ssid < 0 || ssid > 15) {
                        printf("SSID must follow '-' and be numeric in the range 0-15 - '%s'\n", name);
                        exit(5);
                }
        }

        buf[6] = ((ssid + '0') << 1) & 0x1E;

        return 0;
}

/*
 * APRS-IS server connection
 *
 */

int aprs_is_connect(char *server, int port, char *aprs_payload)
{
	int i, sockfd, bytes, ret;
        struct hostent *he;
	struct sockaddr_in saddr;

	fprintf(stderr, "Connecting to APRS-IS server %s:%d ... ", server, port);

        if (!(he = gethostbyname(server))) {

                perror("gethostbyname");
                exit(1);
        }

        memset(&saddr, 0, sizeof(saddr));

        saddr.sin_family = AF_INET;
        saddr.sin_addr = *(struct in_addr *)(he->h_addr);
        saddr.sin_port = htons(port);

	sockfd = socket(AF_INET, SOCK_STREAM, 0);

	if (connect(sockfd, (struct sockaddr *) &saddr, sizeof(saddr)) < 0) {

		perror("connect");
		exit(1);
	}

	if (verbose)
		fprintf(stderr, "connected\n");

	/* talk to server */
	
	if (verbose)
		fprintf(stderr, "sending aprs-is login string");

	bytes = snprintf(send_buf, sizeof(send_buf), "user %s pass %s vers UXmit 0.6\n", mycall, passcode);

	ret = send(sockfd, send_buf, bytes, 0);

	/* check for login success */

	for (i = 0; i < 5; i++) {

        	bytes = recv(sockfd, recv_buf, sizeof(recv_buf), 0);

        	if (bytes < 0) {

			perror("recv");
			exit(1);
		}

		// fprintf(stderr, "read %d bytes: %s\n", bytes, recv_buf);
	}
	
	fprintf(stderr, "Sucess\n");

	bytes = snprintf(send_buf, sizeof(send_buf), "%s\n", aprs_payload);

	ret = send(sockfd, send_buf, bytes, 0);

	sleep(4);

        close(sockfd);

	return 0;
}
